@@ -15,8 +15,8 @@
 static void usage(int status)
 {
     fprintf(stderr,
-	    "usage: setcap [-h] [-q] [-v] [-n <rootid>] (-r|-|<caps>) <filename> "
-	    "[ ... (-r|-|<capsN>) <filenameN> ]\n"
+	    "usage: setcap [--license] [-f] [-h] [-n <rootid>] [-q] [-v]"
+	    " (-r|-|<caps>) <filename> [ ... (-r|-|<capsN>) <filenameN> ]\n"
 	    "\n"
 	    " Note <filename> must be a regular (non-symlink) file.\n"
 	    " -r          remove capability from file\n"
@@ -24,12 +24,12 @@ static void usage(int status)
 	    " <capsN>     cap_from_text(3) formatted file capability\n"
 	    " [ Note: capsh --suggest=\"something...\" might help you pick. ]"
 	    "\n"
-	    " -h          this message and exit status 0\n"
+	    " --license   display the license info\n"
 	    " -f          force setting even when the capability is invalid\n"
+	    " -h          this message and exit status 0\n"
+	    " -n <rootid> write a user namespace (!= 0) limited capability\n"
 	    " -q          quietly\n"
 	    " -v          validate supplied capability matches file\n"
-	    " -n <rootid> write a user namespace (!= 0) limited capability\n"
-	    " --license   display the license info\n"
 	);
     exit(status);
 }
@@ -117,56 +117,57 @@ int main(int argc, char **argv)
     }
 
     cap_t cap_d = NULL;
-    while (--argc > 0) {
+    char **arg = argv+1;
+    for (; --argc > 0; arg++) {
 	const char *text;
 
 	cap_free(cap_d);
 	cap_d = NULL;
 
-	if (!strcmp(*++argv, "-q")) {
-	    quiet = 1;
-	    continue;
-	}
-	if (!strcmp("--license", *argv)) {
+	if (!strcmp("--license", *arg)) {
 	    printf(
 		"%s see LICENSE file for details.\n"
 		"Copyright (c) 1997,2007-8,2020-21 Andrew G. Morgan"
 		" <morgan@kernel.org>\n", argv[0]);
 	    exit(0);
 	}
-	if (!strcmp(*argv, "-f")) {
+	if (!strcmp(*arg, "-f")) {
 	    forced = 1;
 	    continue;
 	}
-	if (!strcmp(*argv, "-h")) {
+	if (!strcmp(*arg, "-h")) {
 	    usage(0);
 	}
-	if (!strcmp(*argv, "-v")) {
-	    verify = 1;
-	    continue;
-	}
-	if (!strcmp(*argv, "-n")) {
+	if (!strcmp(*arg, "-n")) {
 	    if (argc < 2) {
 		fprintf(stderr,
 			"usage: .. -n <rootid> .. - rootid!=0 file caps");
 		exit(1);
 	    }
 	    --argc;
-	    rootid = (uid_t) pos_uint(*++argv, "bad ns rootid", NULL);
+	    rootid = (uid_t) pos_uint(*++arg, "bad ns rootid", NULL);
+	    continue;
+	}
+	if (!strcmp(*arg, "-q")) {
+	    quiet = 1;
+	    continue;
+	}
+	if (!strcmp(*arg, "-v")) {
+	    verify = 1;
 	    continue;
 	}
 
-	if (!strcmp(*argv, "-r")) {
+	if (!strcmp(*arg, "-r")) {
 	    cap_free(cap_d);
 	    cap_d = NULL;
 	} else {
-	    if (!strcmp(*argv,"-")) {
-		retval = read_caps(quiet, *argv, buffer);
+	    if (!strcmp(*arg,"-")) {
+		retval = read_caps(quiet, *arg, buffer);
 		if (retval)
 		    usage(1);
 		text = buffer;
 	    } else {
-		text = *argv;
+		text = *arg;
 	    }
 
 	    int non_space = 0, j;
@@ -182,6 +183,7 @@ int main(int argc, char **argv)
 	    }
 	    cap_d = cap_from_text(text);
 	    if (cap_d == NULL) {
+		printf("argument: %s\n", text);
 		perror("fatal error");
 		usage(1);
 	    }
@@ -198,8 +200,10 @@ int main(int argc, char **argv)
 #endif
 	}
 
-	if (--argc <= 0)
+	if (--argc <= 0) {
 	    usage(1);
+	}
+
 	/*
 	 * Set the filesystem capability for this file.
 	 */
@@ -215,7 +219,7 @@ int main(int argc, char **argv)
 		}
 	    }
 
-	    cap_on_file = cap_get_file(*++argv);
+	    cap_on_file = cap_get_file(*++arg);
 	    if (cap_on_file == NULL) {
 		cap_on_file = cap_init();
 		if (cap_on_file == NULL) {
@@ -233,7 +237,7 @@ int main(int argc, char **argv)
 		    if (rootid != f_rootid) {
 			printf("nsowner[got=%d, want=%d],", f_rootid, rootid);
 		    }
-		    printf("%s differs in [%s%s%s]\n", *argv,
+		    printf("%s differs in [%s%s%s]\n", *arg,
 			   CAP_DIFFERS(cmp, CAP_PERMITTED) ? "p" : "",
 			   CAP_DIFFERS(cmp, CAP_INHERITABLE) ? "i" : "",
 			   CAP_DIFFERS(cmp, CAP_EFFECTIVE) ? "e" : "");
@@ -241,7 +245,7 @@ int main(int argc, char **argv)
 		exit(1);
 	    }
 	    if (!quiet) {
-		printf("%s: OK\n", *argv);
+		printf("%s: OK\n", *arg);
 	    }
 	} else {
 	    if (!tried_to_cap_setfcap) {
@@ -294,19 +298,19 @@ int main(int argc, char **argv)
 	    }
 #endif /* def linux */
 	    errno = 0;
-	    retval = cap_set_file(*++argv, cap_d);
+	    retval = cap_set_file(*++arg, cap_d);
 	    if (retval != 0) {
 		switch (errno) {
 		case EINVAL:
 		    fprintf(stderr,
 			    "Invalid file '%s' for capability operation\n",
-			    argv[0]);
+			    *arg);
 		    exit(1);
 		case ENODATA:
 		    if (cap_d == NULL) {
 			fprintf(stderr,
 				"File '%s' has no capablity to remove\n",
-				argv[0]);
+				*arg);
 			if (forced) {
 			    break;
 			}
@@ -316,7 +320,7 @@ int main(int argc, char **argv)
 		default:
 		    fprintf(stderr,
 			    "Failed to set capabilities on file '%s': %s\n",
-			    argv[0], strerror(errno));
+			    *arg, strerror(errno));
 		    exit(1);
 		}
 	    }
